@@ -19,8 +19,8 @@ from sbi.inference.base import (  # noqa: F401
     infer,
     simulate_for_sbi,
 )
+from sbi.inference.snle.mnle import MNLE
 from sbi.inference.snle.snle_a import SNLE_A
-# Unimplemented: don't export
 from sbi.inference.snpe.snpe_a import SNPE_A
 from sbi.inference.snpe.snpe_b import SNPE_B
 from sbi.inference.snpe.snpe_c import SNPE_C  # noqa: F401
@@ -32,7 +32,7 @@ _snle_family = ["SNL"]
 
 
 SNPE = APT = SNPE_C
-_snpe_family = ["SNPE_C", "SNPE", "APT"]
+_snpe_family = ["SNPE_A", "SNPE_C", "SNPE", "APT"]
 
 
 SRE = SNRE_B
@@ -45,3 +45,18 @@ _abc_family = ["ABC", "MCABC", "SMC", "SMCABC"]
 
 
 __all__ = _snpe_family + _snre_family + _snle_family + _abc_family
+
+from sbi.inference.posteriors.direct_posterior import DirectPosterior
+from sbi.inference.posteriors.mcmc_posterior import MCMCPosterior
+from sbi.inference.posteriors.rejection_posterior import RejectionPosterior
+from sbi.inference.posteriors.vi_posterior import VIPosterior
+from sbi.inference.potentials.likelihood_based_potential import (
+    likelihood_estimator_based_potential,
+    mixed_likelihood_estimator_based_potential,
+)
+from sbi.inference.potentials.posterior_based_potential import (
+    posterior_estimator_based_potential,
+)
+from sbi.inference.potentials.ratio_based_potential import (
+    ratio_estimator_based_potential,
+)
