@@ -3,7 +3,6 @@ from typing import Any, Callable, Dict, Optional, Tuple, Union
 import torch
 from numpy import ndarray
 from torch import Tensor
-from torch.distributions.transforms import Transform
 
 from sbi.inference.abc.abc_base import ABCBASE
 from sbi.utils import KDEWrapper, get_kde, process_x
@@ -19,7 +18,7 @@ class MCABC(ABCBASE):
         simulation_batch_size: int = 1,
         show_progress_bars: bool = True,
     ):
-        """Monte-Carlo Approximate Bayesian Computation (Rejection ABC) [1].
+        r"""Monte-Carlo Approximate Bayesian Computation (Rejection ABC) [1].
 
         [1] Pritchard, J. K., Seielstad, M. T., Perez-Lezaun, A., & Feldman, M. W.
         (1999). Population growth of human Y chromosomes: a study of Y chromosome
@@ -65,7 +64,7 @@ class MCABC(ABCBASE):
         sass_fraction: float = 0.25,
         sass_expansion_degree: int = 1,
         kde: bool = False,
-        kde_kwargs: Optional[Dict[str, Any]] = {},
+        kde_kwargs: Dict[str, Any] = {},
         return_summary: bool = False,
     ) -> Union[Tuple[Tensor, dict], Tuple[KDEWrapper, dict], Tensor, KDEWrapper]:
         r"""Run MCABC and return accepted parameters or KDE object fitted on them.
