@@ -5,18 +5,21 @@ from abc import ABC, abstractmethod
 from copy import deepcopy
 from datetime import datetime
 from pathlib import Path
-from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast
+from typing import Any, Callable, Dict, Optional, Tuple, Union
 from warnings import warn
 
 import torch
 from torch import Tensor
+from torch.distributions import Distribution
 from torch.utils import data
-from torch.utils.tensorboard import SummaryWriter
+from torch.utils.data.sampler import SubsetRandomSampler
+from torch.utils.tensorboard.writer import SummaryWriter
 
 import sbi.inference
 from sbi.inference.posteriors.base_posterior import NeuralPosterior
 from sbi.simulators.simutils import simulate_in_batches
 from sbi.utils import (
+    check_prior,
     get_log_root,
     handle_invalid_x,
     warn_if_zscoring_changes_data,
@@ -24,15 +27,18 @@ from sbi.utils import (
     warn_on_invalid_x_for_snpec_leakage,
 )
 from sbi.utils.sbiutils import get_simulations_since_round
-from sbi.utils.torchutils import process_device
+from sbi.utils.torchutils import check_if_prior_on_device, process_device
 from sbi.utils.user_input_checks import prepare_for_sbi
 
 
 def infer(
-    simulator: Callable, prior, method: str, num_simulations: int, num_workers: int = 1
+    simulator: Callable,
+    prior: Distribution,
+    method: str,
+    num_simulations: int,
+    num_workers: int = 1,
 ) -> NeuralPosterior:
-    r"""
-    Return posterior distribution by running simulation-based inference.
+    r"""Runs simulation-based inference and returns the posterior.
 
     This function provides a simple interface to run sbi. Inference is run for a single
     round and hence the returned posterior $p(\theta|x)$ can be sampled and evaluated
@@ -69,7 +75,7 @@ def infer(
 
     simulator, prior = prepare_for_sbi(simulator, prior)
 
-    inference = method_fun(prior)
+    inference = method_fun(prior=prior)
     theta, x = simulate_for_sbi(
         simulator=simulator,
         proposal=prior,
@@ -87,21 +93,18 @@ class NeuralInference(ABC):
 
     def __init__(
         self,
-        prior: Optional[Any] = None,
+        prior: Optional[Distribution] = None,
         device: str = "cpu",
         logging_level: Union[int, str] = "WARNING",
         summary_writer: Optional[SummaryWriter] = None,
         show_progress_bars: bool = True,
-        **unused_args,
     ):
-        r"""
-        Base class for inference methods.
+        r"""Base class for inference methods.
 
         Args:
             prior: A probability distribution that expresses prior knowledge about the
-                parameters, e.g. which ranges are meaningful for them. Any
-                object with `.log_prob()`and `.sample()` (for example, a PyTorch
-                distribution) can be used.
+                parameters, e.g. which ranges are meaningful for them. Must be a PyTorch
+                distribution, see FAQ for details on how to use custom distributions.
             device: torch device on which to train the neural net and on which to
                 perform all posterior operations, e.g. gpu or cpu.
             logging_level: Minimum severity of messages to log. One of the strings
@@ -110,26 +113,13 @@ class NeuralInference(ABC):
                 file location (default is `<current working directory>/logs`.)
             show_progress_bars: Whether to show a progressbar during simulation and
                 sampling.
-            unused_args: Absorbs additional arguments. No entries will be used. If it
-                is not empty, we warn. In future versions, when the new interface of
-                0.14.0 is more mature, we will remove this argument.
         """
 
-        self._device = process_device(device, prior=prior)
-
-        if unused_args:
-            warn(
-                f"You passed some keyword arguments that will not be used. "
-                f"Specifically, the unused arguments are: {list(unused_args.keys())}. "
-                f"These arguments might have been supported in sbi "
-                f"versions <0.14.0. Since 0.14.0, the API was changed. Please consult "
-                f"the corresponding pull request on github: "
-                f"https://github.com/mackelab/sbi/pull/378 and tutorials: "
-                f"https://www.mackelab.org/sbi/tutorial/02_flexible_interface/ for "
-                f"further information.",
-            )
-
+        self._device = process_device(device)
+        check_prior(prior)
+        check_if_prior_on_device(self._device, prior)
         self._prior = prior
+
         self._posterior = None
         self._neural_net = None
         self._x_shape = None
@@ -145,6 +135,7 @@ class NeuralInference(ABC):
         self._data_round_index = []
 
         self._round = 0
+        self._val_log_prob = float("-Inf")
 
         # XXX We could instantiate here the Posterior for all children. Two problems:
         #     1. We must dispatch to right PotentialProvider for mcmc based on name
@@ -165,66 +156,13 @@ class NeuralInference(ABC):
             epoch_durations_sec=[],
         )
 
-    def __call__(self, unused_args):
-        """
-        f"Deprecated. The inference object is no longer callable as of `sbi` v0.14.0. "
-        f"Please consult our website for a tutorial on how to adapt your code: "
-        f"https://www.mackelab.org/sbi/tutorial/02_flexible_interface/ . For "
-        f"further information, see the corresponding pull request on github:
-        f"https://github.com/mackelab/sbi/pull/378.",
-        """
-        raise NameError(
-            f"The inference object is no longer callable as of `sbi` v0.14.0. "
-            f"Please consult the release notes for a tutorial on how to adapt your "
-            f"code: https://github.com/mackelab/sbi/releases/tag/v0.14.0"
-            f"For more information, visit our website: "
-            f"https://www.mackelab.org/sbi/tutorial/02_flexible_interface/ or "
-            f"see the corresponding pull request on github: "
-            f"https://github.com/mackelab/sbi/pull/378.",
-        )
-
-    def provide_presimulated(
-        self, theta: Tensor, x: Tensor, from_round: int = 0
-    ) -> None:
-        r"""
-        Deprecated since sbi 0.14.0.
-
-        Instead of using this, please use `.append_simulations()`. Please consult
-        release notes to see how you can update your code:
-        https://github.com/mackelab/sbi/releases/tag/v0.14.0
-        More information can be found under the corresponding pull request on github:
-        https://github.com/mackelab/sbi/pull/378
-        and tutorials:
-        https://www.mackelab.org/sbi/tutorial/02_flexible_interface/
-
-        Provide external $\theta$ and $x$ to be used for training later on.
-
-        Args:
-            theta: Parameter sets used to generate presimulated data.
-            x: Simulation outputs of presimulated data.
-            from_round: Which round the data was simulated from. `from_round=0` means
-                that the data came from the first round, i.e. the prior.
-        """
-        raise NameError(
-            f"Deprecated since sbi 0.14.0. "
-            f"Instead of using this, please use `.append_simulations()`. Please "
-            f"consult release notes to see how you can update your code: "
-            f"https://github.com/mackelab/sbi/releases/tag/v0.14.0"
-            f"More information can be found under the corresponding pull request on "
-            f"github: "
-            f"https://github.com/mackelab/sbi/pull/378"
-            f"and tutorials: "
-            f"https://www.mackelab.org/sbi/tutorial/02_flexible_interface/",
-        )
-
     def get_simulations(
         self,
         starting_round: int = 0,
         exclude_invalid_x: bool = True,
         warn_on_invalid: bool = True,
     ) -> Tuple[Tensor, Tensor, Tensor]:
-        r"""
-        Returns all $\theta$, $x$, and prior_masks from rounds >= `starting_round`.
+        r"""Returns all $\theta$, $x$, and prior_masks from rounds >= `starting_round`.
 
         If requested, do not return invalid data.
 
@@ -252,7 +190,7 @@ class NeuralInference(ABC):
         # Check for NaNs in simulations.
         is_valid_x, num_nans, num_infs = handle_invalid_x(x, exclude_invalid_x)
         # Check for problematic z-scoring
-        warn_if_zscoring_changes_data(x)
+        warn_if_zscoring_changes_data(x[is_valid_x])
         if warn_on_invalid:
             warn_on_invalid_x(num_nans, num_infs, exclude_invalid_x)
             warn_on_invalid_x_for_snpec_leakage(
@@ -273,7 +211,7 @@ class NeuralInference(ABC):
         calibration_kernel: Optional[Callable] = None,
         exclude_invalid_x: bool = True,
         discard_prior_samples: bool = False,
-        retrain_from_scratch_each_round: bool = False,
+        retrain_from_scratch: bool = False,
         show_train_summary: bool = False,
     ) -> NeuralPosterior:
         raise NotImplementedError
@@ -294,7 +232,7 @@ class NeuralInference(ABC):
             resume_training: Whether the current call is resuming training so that no
                 new training and validation indices into the dataset have to be created.
             dataloader_kwargs: Additional or updated kwargs to be passed to the training
-                and validation dataloaders (like, e.g., a collate_fn)
+                and validation dataloaders (like, e.g., a collate_fn).
 
         Returns:
             Tuple of dataloaders for training and validation.
@@ -322,24 +260,18 @@ class NeuralInference(ABC):
         train_loader_kwargs = {
             "batch_size": min(training_batch_size, num_training_examples),
             "drop_last": True,
-            "sampler": data.sampler.SubsetRandomSampler(self.train_indices),
+            "sampler": SubsetRandomSampler(self.train_indices.tolist()),
         }
-        train_loader_kwargs = (
-            dict(train_loader_kwargs, **dataloader_kwargs)
-            if dataloader_kwargs is not None
-            else train_loader_kwargs
-        )
         val_loader_kwargs = {
             "batch_size": min(training_batch_size, num_validation_examples),
             "shuffle": False,
             "drop_last": True,
-            "sampler": data.sampler.SubsetRandomSampler(self.val_indices),
+            "sampler": SubsetRandomSampler(self.val_indices.tolist()),
         }
-        val_loader_kwargs = (
-            dict(val_loader_kwargs, **dataloader_kwargs)
-            if dataloader_kwargs is not None
-            else val_loader_kwargs
-        )
+        if dataloader_kwargs is not None:
+            train_loader_kwargs = dict(train_loader_kwargs, **dataloader_kwargs)
+            val_loader_kwargs = dict(val_loader_kwargs, **dataloader_kwargs)
+
         train_loader = data.DataLoader(dataset, **train_loader_kwargs)
         val_loader = data.DataLoader(dataset, **val_loader_kwargs)
 
@@ -359,6 +291,7 @@ class NeuralInference(ABC):
         """
         converged = False
 
+        assert self._neural_net is not None
         neural_net = self._neural_net
 
         # (Re)-start the epoch count with the first epoch or any improvement.
@@ -385,21 +318,6 @@ class NeuralInference(ABC):
         )
         return SummaryWriter(logdir)
 
-    @staticmethod
-    def _ensure_list(
-        num_simulations_per_round: Union[List[int], int], num_rounds: int
-    ) -> List[int]:
-        """Return `num_simulations_per_round` as a list of length `num_rounds`."""
-        try:
-            assert len(num_simulations_per_round) == num_rounds, (
-                "Please provide a list with number of simulations per round for each "
-                "round, or a single integer to be used for all rounds."
-            )
-        except TypeError:
-            num_simulations_per_round: List = [num_simulations_per_round] * num_rounds
-
-        return cast(list, num_simulations_per_round)
-
     @staticmethod
     def _describe_round(round_: int, summary: Dict[str, list]) -> str:
         epochs = summary["epochs"][-1]
@@ -417,17 +335,22 @@ class NeuralInference(ABC):
         return description
 
     @staticmethod
-    def _maybe_show_progress(show=bool, epoch=int) -> None:
+    def _maybe_show_progress(show: bool, epoch: int) -> None:
         if show:
             # end="\r" deletes the print statement when a new one appears.
-            # https://stackoverflow.com/questions/3419984/
-            print("Training neural network. Epochs trained: ", epoch, end="\r")
+            # https://stackoverflow.com/questions/3419984/. `\r` in the beginning due
+            # to #330.
+            print("\r", f"Training neural network. Epochs trained: {epoch}", end="")
 
     def _report_convergence_at_end(
         self, epoch: int, stop_after_epochs: int, max_num_epochs: int
     ) -> None:
         if self._converged(epoch, stop_after_epochs):
-            print(f"Neural network successfully converged after {epoch} epochs.")
+            print(
+                "\r",
+                f"Neural network successfully converged after {epoch} epochs.",
+                end="",
+            )
         elif max_num_epochs == epoch:
             warn(
                 "Maximum number of epochs `max_num_epochs={max_num_epochs}` reached,"
@@ -441,6 +364,14 @@ class NeuralInference(ABC):
 
         Statistics are extracted from the arguments and from entries in self._summary
         created during training.
+
+        Scalar tags:
+            - median_observation_distances
+            - epochs_trained
+            - best_validation_log_prob
+            - validation_log_probs_across_rounds
+            - train_log_probs_across_rounds
+            - epoch_durations_sec_across_rounds
         """
 
         # NB. This is a subset of the logging as done in `GH:conormdurkan/lfi`. A big
@@ -477,7 +408,9 @@ class NeuralInference(ABC):
 
         # Add validation log prob for every epoch.
         # Offset with all previous epochs.
-        offset = torch.tensor(self._summary["epochs"][:-1], dtype=int).sum().item()
+        offset = (
+            torch.tensor(self._summary["epochs"][:-1], dtype=torch.int).sum().item()
+        )
         for i, vlp in enumerate(self._summary["validation_log_probs"][offset:]):
             self._summary_writer.add_scalar(
                 tag="validation_log_probs_across_rounds",
@@ -505,6 +438,42 @@ class NeuralInference(ABC):
     def summary(self):
         return self._summary
 
+    def __getstate__(self) -> Dict:
+        """Returns the state of the object that is supposed to be pickled.
+
+        Attributes that can not be serialized are set to `None`.
+
+        Returns:
+            Dictionary containing the state.
+        """
+        warn(
+            "When the inference object is pickled, the behaviour of the loaded object "
+            "changes in the following two ways: "
+            "1) `.train(..., retrain_from_scratch=True)` is not supported. "
+            "2) When the loaded object calls the `.train()` method, it generates a new "
+            "tensorboard summary writer (instead of appending to the current one)."
+        )
+        dict_to_save = {}
+        unpicklable_attributes = ["_summary_writer", "_build_neural_net"]
+        for key in self.__dict__.keys():
+            if key in unpicklable_attributes:
+                dict_to_save[key] = None
+            else:
+                dict_to_save[key] = self.__dict__[key]
+        return dict_to_save
+
+    def __setstate__(self, state_dict: Dict):
+        """Sets the state when being loaded from pickle.
+
+        Also creates a new summary writer (because the previous one was set to `None`
+        during serializing, see `__get_state__()`).
+
+        Args:
+            state_dict: State to be restored.
+        """
+        state_dict["_summary_writer"] = self._default_summary_writer()
+        self.__dict__ = state_dict
+
 
 def simulate_for_sbi(
     simulator: Callable,
@@ -514,8 +483,7 @@ def simulate_for_sbi(
     simulation_batch_size: int = 1,
     show_progress_bar: bool = True,
 ) -> Tuple[Tensor, Tensor]:
-    r"""
-    Returns ($\theta, x$) pairs obtained from sampling the proposal and simulating.
+    r"""Returns ($\theta, x$) pairs obtained from sampling the proposal and simulating.
 
     This function performs two steps:
 
@@ -542,8 +510,6 @@ def simulate_for_sbi(
     Returns: Sampled parameters $\theta$ and simulation-outputs $x$.
     """
 
-    check_if_proposal_has_default_x(proposal)
-
     theta = proposal.sample((num_simulations,))
 
     x = simulate_in_batches(
@@ -554,8 +520,7 @@ def simulate_for_sbi(
 
 
 def check_if_proposal_has_default_x(proposal: Any):
-    """
-    Check for validity of the provided proposal distribution.
+    """Check for validity of the provided proposal distribution.
 
     If the proposal is a `NeuralPosterior`, we check if the default_x is set and
     if it matches the `_x_o_training_focused_on`.
