@@ -7,15 +7,22 @@ from typing import Any, Callable, Dict, Optional, Union
 from warnings import warn
 
 import torch
-from torch import Tensor, ones, optim
-from torch.nn.utils import clip_grad_norm_
+from torch import Tensor, nn, ones, optim
+from torch.distributions import Distribution
+from torch.nn.utils.clip_grad import clip_grad_norm_
 from torch.utils import data
-from torch.utils.tensorboard import SummaryWriter
+from torch.utils.tensorboard.writer import SummaryWriter
 
 from sbi import utils as utils
 from sbi.inference import NeuralInference, check_if_proposal_has_default_x
-from sbi.inference.posteriors.direct_posterior import DirectPosterior
-from sbi.types import TorchModule
+from sbi.inference.posteriors import (
+    DirectPosterior,
+    MCMCPosterior,
+    RejectionPosterior,
+    VIPosterior,
+)
+from sbi.inference.posteriors.base_posterior import NeuralPosterior
+from sbi.inference.potentials import posterior_estimator_based_potential
 from sbi.utils import (
     RestrictedPrior,
     check_estimator_arg,
@@ -29,13 +36,12 @@ from sbi.utils.sbiutils import ImproperEmpirical, mask_sims_from_prior
 class PosteriorEstimator(NeuralInference, ABC):
     def __init__(
         self,
-        prior: Optional[Any] = None,
+        prior: Optional[Distribution] = None,
         density_estimator: Union[str, Callable] = "maf",
         device: str = "cpu",
         logging_level: Union[int, str] = "WARNING",
         summary_writer: Optional[SummaryWriter] = None,
         show_progress_bars: bool = True,
-        **unused_args,
     ):
         """Base class for Sequential Neural Posterior Estimation methods.
 
@@ -48,9 +54,6 @@ class PosteriorEstimator(NeuralInference, ABC):
                 needs to return a PyTorch `nn.Module` implementing the density
                 estimator. The density estimator needs to provide the methods
                 `.log_prob` and `.sample()`.
-            unused_args: Absorbs additional arguments. No entries will be used. If it
-                is not empty, we warn. In future versions, when the new interface of
-                0.14.0 is more mature, we will remove this argument.
 
         See docstring of `NeuralInference` class for all other arguments.
         """
@@ -61,7 +64,6 @@ class PosteriorEstimator(NeuralInference, ABC):
             logging_level=logging_level,
             summary_writer=summary_writer,
             show_progress_bars=show_progress_bars,
-            **unused_args,
         )
 
         # As detailed in the docstring, `density_estimator` is either a string or
@@ -82,10 +84,12 @@ class PosteriorEstimator(NeuralInference, ABC):
         self._summary.update({"rejection_sampling_acceptance_rates": []})  # type:ignore
 
     def append_simulations(
-        self, theta: Tensor, x: Tensor, proposal: Optional[Any] = None
+        self,
+        theta: Tensor,
+        x: Tensor,
+        proposal: Optional[DirectPosterior] = None,
     ) -> "PosteriorEstimator":
-        r"""
-        Store parameters and simulation outputs to use them for later training.
+        r"""Store parameters and simulation outputs to use them for later training.
 
         Data are stored as entries in lists for each type of variable (parameter/data).
 
@@ -154,18 +158,17 @@ class PosteriorEstimator(NeuralInference, ABC):
         learning_rate: float = 5e-4,
         validation_fraction: float = 0.1,
         stop_after_epochs: int = 20,
-        max_num_epochs: Optional[int] = None,
+        max_num_epochs: int = 2**31 - 1,
         clip_max_norm: Optional[float] = 5.0,
         calibration_kernel: Optional[Callable] = None,
         exclude_invalid_x: bool = True,
         resume_training: bool = False,
         discard_prior_samples: bool = False,
-        retrain_from_scratch_each_round: bool = False,
+        retrain_from_scratch: bool = False,
         show_train_summary: bool = False,
         dataloader_kwargs: Optional[dict] = None,
-    ) -> DirectPosterior:
-        r"""
-        Return density estimator that approximates the distribution $p(\theta|x)$.
+    ) -> nn.Module:
+        r"""Return density estimator that approximates the distribution $p(\theta|x)$.
 
         Args:
             training_batch_size: Training batch size.
@@ -174,8 +177,8 @@ class PosteriorEstimator(NeuralInference, ABC):
             stop_after_epochs: The number of epochs to wait for improvement on the
                 validation set before terminating training.
             max_num_epochs: Maximum number of epochs to run. If reached, we stop
-                training even when the validation loss is still decreasing. If None, we
-                train until validation loss increases (see also `stop_after_epochs`).
+                training even when the validation loss is still decreasing. Otherwise,
+                we train until validation loss increases (see also `stop_after_epochs`).
             clip_max_norm: Value at which to clip the total gradient norm in order to
                 prevent exploding gradients. Use None for no clipping.
             calibration_kernel: A function to calibrate the loss with respect to the
@@ -189,7 +192,7 @@ class PosteriorEstimator(NeuralInference, ABC):
             discard_prior_samples: Whether to discard samples simulated in round 1, i.e.
                 from the prior. Training may be sped up by ignoring such less targeted
                 samples.
-            retrain_from_scratch_each_round: Whether to retrain the conditional density
+            retrain_from_scratch: Whether to retrain the conditional density
                 estimator for the posterior from scratch each round.
             show_train_summary: Whether to print the number of epochs and validation
                 loss after the training.
@@ -204,8 +207,6 @@ class PosteriorEstimator(NeuralInference, ABC):
         if calibration_kernel is None:
             calibration_kernel = lambda x: ones([len(x)], device=self._device)
 
-        max_num_epochs = 2 ** 31 - 1 if max_num_epochs is None else max_num_epochs
-
         # Starting index for the training set (1 = discard round-0 samples).
         start_idx = int(discard_prior_samples and self._round > 0)
 
@@ -242,7 +243,7 @@ class PosteriorEstimator(NeuralInference, ABC):
         # arguments, which will build the neural network.
         # This is passed into NeuralPosterior, to create a neural posterior which
         # can `sample()` and `log_prob()`. The network is accessible via `.net`.
-        if self._neural_net is None or retrain_from_scratch_each_round:
+        if self._neural_net is None or retrain_from_scratch:
             self._neural_net = self._build_neural_net(
                 theta[self.train_indices], x[self.train_indices]
             )
@@ -271,7 +272,7 @@ class PosteriorEstimator(NeuralInference, ABC):
 
             # Train for a single epoch.
             self._neural_net.train()
-            train_log_prob_sum = 0
+            train_log_probs_sum = 0
             epoch_start_time = time.time()
             for batch in train_loader:
                 self.optimizer.zero_grad()
@@ -282,15 +283,13 @@ class PosteriorEstimator(NeuralInference, ABC):
                     batch[2].to(self._device),
                 )
 
-                batch_loss = torch.mean(
-                    self._loss(
-                        theta_batch, x_batch, masks_batch, proposal, calibration_kernel
-                    )
+                train_losses = self._loss(
+                    theta_batch, x_batch, masks_batch, proposal, calibration_kernel
                 )
+                train_loss = torch.mean(train_losses)
+                train_log_probs_sum -= train_losses.sum().item()
 
-                train_log_prob_sum += batch_loss.sum().item()
-
-                batch_loss.backward()
+                train_loss.backward()
                 if clip_max_norm is not None:
                     clip_grad_norm_(
                         self._neural_net.parameters(), max_norm=clip_max_norm
@@ -299,12 +298,14 @@ class PosteriorEstimator(NeuralInference, ABC):
 
             self.epoch += 1
 
-            train_log_prob_sum /= int(theta.shape[0] * (1.0 - validation_fraction))
-            self._summary["train_log_probs"].append(train_log_prob_sum)
+            train_log_prob_average = train_log_probs_sum / (
+                len(train_loader) * train_loader.batch_size  # type: ignore
+            )
+            self._summary["train_log_probs"].append(train_log_prob_average)
 
             # Calculate validation performance.
             self._neural_net.eval()
-            log_prob_sum = 0
+            val_log_prob_sum = 0
 
             with torch.no_grad():
                 for batch in val_loader:
@@ -314,14 +315,18 @@ class PosteriorEstimator(NeuralInference, ABC):
                         batch[2].to(self._device),
                     )
                     # Take negative loss here to get validation log_prob.
-                    batch_log_prob = -self._loss(
-                        theta_batch, x_batch, masks_batch, proposal, calibration_kernel
+                    val_losses = self._loss(
+                        theta_batch,
+                        x_batch,
+                        masks_batch,
+                        proposal,
+                        calibration_kernel,
                     )
-                    log_prob_sum += batch_log_prob.sum().item()
+                    val_log_prob_sum -= val_losses.sum().item()
 
             # Take mean over all validation samples.
-            self._val_log_prob = log_prob_sum / (
-                len(val_loader) * val_loader.batch_size
+            self._val_log_prob = val_log_prob_sum / (
+                len(val_loader) * val_loader.batch_size  # type: ignore
             )
             # Log validation log prob for every epoch.
             self._summary["validation_log_probs"].append(self._val_log_prob)
@@ -342,23 +347,27 @@ class PosteriorEstimator(NeuralInference, ABC):
         if show_train_summary:
             print(self._describe_round(self._round, self._summary))
 
+        # Avoid keeping the gradients in the resulting network, which can
+        # cause memory leakage when benchmarking.
+        self._neural_net.zero_grad(set_to_none=True)
+
         return deepcopy(self._neural_net)
 
     def build_posterior(
         self,
-        density_estimator: Optional[TorchModule] = None,
+        density_estimator: Optional[nn.Module] = None,
+        prior: Optional[Distribution] = None,
         sample_with: str = "rejection",
         mcmc_method: str = "slice_np",
-        mcmc_parameters: Optional[Dict[str, Any]] = None,
-        rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
-        sample_with_mcmc: Optional[bool] = None,
-    ) -> DirectPosterior:
-        r"""
-        Build posterior from the neural density estimator.
+        vi_method: str = "rKL",
+        mcmc_parameters: Dict[str, Any] = {},
+        vi_parameters: Dict[str, Any] = {},
+        rejection_sampling_parameters: Dict[str, Any] = {},
+    ) -> Union[MCMCPosterior, RejectionPosterior, VIPosterior, DirectPosterior]:
+        r"""Build posterior from the neural density estimator.
 
         For SNPE, the posterior distribution that is returned here implements the
         following functionality over the raw neural density estimator:
-
         - correct the calculation of the log probability such that it compensates for
             the leakage.
         - reject samples that lie outside of the prior bounds.
@@ -368,77 +377,100 @@ class PosteriorEstimator(NeuralInference, ABC):
         Args:
             density_estimator: The density estimator that the posterior is based on.
                 If `None`, use the latest neural density estimator that was trained.
+            prior: Prior distribution.
             sample_with: Method to use for sampling from the posterior. Must be one of
-                [`rejection` | `mcmc`]. With default parameters, `rejection` samples
-                from the posterior estimated by the neural net and rejects only if the
-                samples are outside of the prior support.
+                [`mcmc` | `rejection` | `vi`].
             mcmc_method: Method used for MCMC sampling, one of `slice_np`, `slice`,
                 `hmc`, `nuts`. Currently defaults to `slice_np` for a custom numpy
                 implementation of slice sampling; select `hmc`, `nuts` or `slice` for
                 Pyro-based sampling.
-            mcmc_parameters: Dictionary overriding the default parameters for MCMC.
-                The following parameters are supported: `thin` to set the thinning
-                factor for the chain, `warmup_steps` to set the initial number of
-                samples to discard, `num_chains` for the number of chains,
-                `init_strategy` for the initialisation strategy for chains; `prior` will
-                draw init locations from prior, whereas `sir` will use
-                Sequential-Importance-Resampling using `init_strategy_num_candidates`
-                to find init locations.
-            rejection_sampling_parameters: Dictionary overriding the default parameters
-                for rejection sampling. The following parameters are supported:
-                `proposal` as the proposal distribtution (default is the trained
-                neural net). `max_sampling_batch_size` as the batchsize of samples
-                being drawn from the proposal at every iteration.
-                `num_samples_to_find_max` as the number of samples that are used to
-                find the maximum of the `potential_fn / proposal` ratio.
-                `num_iter_to_find_max` as the number of gradient ascent iterations to
-                find the maximum of that ratio. `m` as multiplier to that ratio.
-            sample_with_mcmc: Deprecated since `sbi v0.16.0`. Use `sample_with=mcmc`
-                instead.
+            vi_method: Method used for VI, one of [`rKL`, `fKL`, `IW`, `alpha`]. Note
+                some of the methods admit a `mode seeking` property (e.g. rKL) whereas
+                some admit a `mass covering` one (e.g fKL).
+            mcmc_parameters: Additional kwargs passed to `MCMCPosterior`.
+            vi_parameters: Additional kwargs passed to `VIPosterior`.
+            rejection_sampling_parameters: Additional kwargs passed to
+                `RejectionPosterior` or `DirectPosterior`. By default,
+                `DirectPosterior` is used. Only if `rejection_sampling_parameters`
+                contains `proposal`, a `RejectionPosterior` is instantiated.
 
         Returns:
-            Posterior $p(\theta|x)$  with `.sample()` and `.log_prob()` methods.
+            Posterior $p(\theta|x)$  with `.sample()` and `.log_prob()` methods
+            (the returned log-probability is unnormalized).
         """
-        if sample_with_mcmc is not None:
-            warn(
-                f"You set `sample_with_mcmc={sample_with_mcmc}`. This is deprecated "
-                "since `sbi v0.17.0` and will lead to an error in future versions. "
-                "Please use `sample_with='mcmc'` instead."
+        if prior is None:
+            assert self._prior is not None, (
+                "You did not pass a prior. You have to pass the prior either at "
+                "initialization `inference = SNPE(prior)` or to "
+                "`.build_posterior(prior=prior)`."
             )
-            if sample_with_mcmc:
-                sample_with = "mcmc"
+            prior = self._prior
+        else:
+            utils.check_prior(prior)
 
         if density_estimator is None:
-            density_estimator = self._neural_net
+            posterior_estimator = self._neural_net
             # If internal net is used device is defined.
             device = self._device
         else:
+            posterior_estimator = density_estimator
             # Otherwise, infer it from the device of the net parameters.
             device = next(density_estimator.parameters()).device.type
 
-        self._posterior = DirectPosterior(
-            method_family="snpe",
-            neural_net=density_estimator,
-            prior=self._prior,
-            x_shape=self._x_shape,
-            sample_with=sample_with,
-            mcmc_method=mcmc_method,
-            mcmc_parameters=mcmc_parameters,
-            rejection_sampling_parameters=rejection_sampling_parameters,
-            device=device,
+        potential_fn, theta_transform = posterior_estimator_based_potential(
+            posterior_estimator=posterior_estimator, prior=prior, x_o=None
         )
 
-        self._posterior._num_trained_rounds = self._round + 1
+        if sample_with == "rejection":
+            if "proposal" in rejection_sampling_parameters.keys():
+                self._posterior = RejectionPosterior(
+                    potential_fn=potential_fn,
+                    device=device,
+                    x_shape=self._x_shape,
+                    **rejection_sampling_parameters,
+                )
+            else:
+                self._posterior = DirectPosterior(
+                    posterior_estimator=posterior_estimator,
+                    prior=prior,
+                    x_shape=self._x_shape,
+                    device=device,
+                )
+        elif sample_with == "mcmc":
+            self._posterior = MCMCPosterior(
+                potential_fn=potential_fn,
+                theta_transform=theta_transform,
+                proposal=prior,
+                method=mcmc_method,
+                device=device,
+                x_shape=self._x_shape,
+                **mcmc_parameters,
+            )
+        elif sample_with == "vi":
+            self._posterior = VIPosterior(
+                potential_fn=potential_fn,
+                theta_transform=theta_transform,
+                prior=prior,  # type: ignore
+                vi_method=vi_method,
+                device=device,
+                x_shape=self._x_shape,
+                **vi_parameters,
+            )
+        else:
+            raise NotImplementedError
 
         # Store models at end of each round.
         self._model_bank.append(deepcopy(self._posterior))
-        self._model_bank[-1].net.eval()
 
         return deepcopy(self._posterior)
 
     @abstractmethod
     def _log_prob_proposal_posterior(
-        self, theta: Tensor, x: Tensor, masks: Tensor, proposal: Optional[Any]
+        self,
+        theta: Tensor,
+        x: Tensor,
+        masks: Tensor,
+        proposal: Optional[Any],
     ) -> Tensor:
         raise NotImplementedError
 
@@ -487,7 +519,7 @@ class PosteriorEstimator(NeuralInference, ABC):
                         "SNPE."
                     )
             elif (
-                not isinstance(proposal, DirectPosterior)
+                not isinstance(proposal, NeuralPosterior)
                 and proposal is not self._prior
             ):
                 warn(
