@@ -3,17 +3,19 @@ from copy import deepcopy
 from typing import Any, Callable, Dict, Optional, Union
 
 import torch
-from torch import Tensor, eye, ones, optim
-from torch.nn.utils import clip_grad_norm_
+from torch import Tensor, eye, nn, ones, optim
+from torch.distributions import Distribution
+from torch.nn.utils.clip_grad import clip_grad_norm_
 from torch.utils import data
-from torch.utils.tensorboard import SummaryWriter
+from torch.utils.tensorboard.writer import SummaryWriter
 
 from sbi import utils as utils
 from sbi.inference.base import NeuralInference
-from sbi.inference.posteriors.ratio_based_posterior import RatioBasedPosterior
-from sbi.types import TorchModule
+from sbi.inference.posteriors import MCMCPosterior, RejectionPosterior, VIPosterior
+from sbi.inference.potentials import ratio_estimator_based_potential
 from sbi.utils import (
     check_estimator_arg,
+    check_prior,
     clamp_and_warn,
     validate_theta_and_x,
     x_shape_from_simulation,
@@ -24,13 +26,12 @@ from sbi.utils.sbiutils import mask_sims_from_prior
 class RatioEstimator(NeuralInference, ABC):
     def __init__(
         self,
-        prior,
+        prior: Optional[Distribution] = None,
         classifier: Union[str, Callable] = "resnet",
         device: str = "cpu",
         logging_level: Union[int, str] = "warning",
         summary_writer: Optional[SummaryWriter] = None,
         show_progress_bars: bool = True,
-        **unused_args
     ):
         r"""Sequential Neural Ratio Estimation.
 
@@ -50,9 +51,6 @@ class RatioEstimator(NeuralInference, ABC):
                 first batch of simulations (theta, x), which can thus be used for shape
                 inference and potentially for z-scoring. It needs to return a PyTorch
                 `nn.Module` implementing the classifier.
-            unused_args: Absorbs additional arguments. No entries will be used. If it
-                is not empty, we warn. In future versions, when the new interface of
-                0.14.0 is more mature, we will remove this argument.
 
         See docstring of `NeuralInference` class for all other arguments.
         """
@@ -63,7 +61,6 @@ class RatioEstimator(NeuralInference, ABC):
             logging_level=logging_level,
             summary_writer=summary_writer,
             show_progress_bars=show_progress_bars,
-            **unused_args
         )
 
         # As detailed in the docstring, `density_estimator` is either a string or
@@ -86,8 +83,7 @@ class RatioEstimator(NeuralInference, ABC):
         x: Tensor,
         from_round: int = 0,
     ) -> "RatioEstimator":
-        r"""
-        Store parameters and simulation outputs to use them for later training.
+        r"""Store parameters and simulation outputs to use them for later training.
 
         Data are stored as entries in lists for each type of variable (parameter/data).
 
@@ -102,6 +98,7 @@ class RatioEstimator(NeuralInference, ABC):
                 With default settings, this is not used at all for `SNRE`. Only when
                 the user later on requests `.train(discard_prior_samples=True)`, we
                 use these indices to find which training data stemmed from the prior.
+
         Returns:
             NeuralInference object (returned so that this function is chainable).
         """
@@ -122,17 +119,16 @@ class RatioEstimator(NeuralInference, ABC):
         learning_rate: float = 5e-4,
         validation_fraction: float = 0.1,
         stop_after_epochs: int = 20,
-        max_num_epochs: Optional[int] = None,
+        max_num_epochs: int = 2**31 - 1,
         clip_max_norm: Optional[float] = 5.0,
         exclude_invalid_x: bool = True,
         resume_training: bool = False,
         discard_prior_samples: bool = False,
-        retrain_from_scratch_each_round: bool = False,
+        retrain_from_scratch: bool = False,
         show_train_summary: bool = False,
         dataloader_kwargs: Optional[Dict] = None,
-    ) -> RatioBasedPosterior:
-        r"""
-        Return classifier that approximates the ratio $p(\theta,x)/p(\theta)p(x)$.
+    ) -> nn.Module:
+        r"""Return classifier that approximates the ratio $p(\theta,x)/p(\theta)p(x)$.
 
         Args:
             num_atoms: Number of atoms to use for classification.
@@ -145,17 +141,15 @@ class RatioEstimator(NeuralInference, ABC):
             discard_prior_samples: Whether to discard samples simulated in round 1, i.e.
                 from the prior. Training may be sped up by ignoring such less targeted
                 samples.
-            retrain_from_scratch_each_round: Whether to retrain the conditional density
+            retrain_from_scratch: Whether to retrain the conditional density
                 estimator for the posterior from scratch each round.
             dataloader_kwargs: Additional or updated kwargs to be passed to the training
-                and validation dataloaders (like, e.g., a collate_fn)
+                and validation dataloaders (like, e.g., a collate_fn).
 
         Returns:
             Classifier that approximates the ratio $p(\theta,x)/p(\theta)p(x)$.
         """
 
-        max_num_epochs = 2 ** 31 - 1 if max_num_epochs is None else max_num_epochs
-
         # Starting index for the training set (1 = discard round-0 samples).
         start_idx = int(discard_prior_samples and self._round > 0)
         # Load data from most recent round.
@@ -177,8 +171,10 @@ class RatioEstimator(NeuralInference, ABC):
 
         clipped_batch_size = min(training_batch_size, len(val_loader))
 
-        num_atoms = clamp_and_warn(
-            "num_atoms", num_atoms, min_val=2, max_val=clipped_batch_size
+        num_atoms = int(
+            clamp_and_warn(
+                "num_atoms", num_atoms, min_val=2, max_val=clipped_batch_size
+            )
         )
 
         # First round or if retraining from scratch:
@@ -186,7 +182,7 @@ class RatioEstimator(NeuralInference, ABC):
         # arguments, which will build the neural network
         # This is passed into NeuralPosterior, to create a neural posterior which
         # can `sample()` and `log_prob()`. The network is accessible via `.net`.
-        if self._neural_net is None or retrain_from_scratch_each_round:
+        if self._neural_net is None or retrain_from_scratch:
             self._neural_net = self._build_neural_net(
                 theta[self.train_indices], x[self.train_indices]
             )
@@ -207,14 +203,19 @@ class RatioEstimator(NeuralInference, ABC):
 
             # Train for a single epoch.
             self._neural_net.train()
+            train_log_probs_sum = 0
             for batch in train_loader:
                 self.optimizer.zero_grad()
                 theta_batch, x_batch = (
                     batch[0].to(self._device),
                     batch[1].to(self._device),
                 )
-                loss = self._loss(theta_batch, x_batch, num_atoms)
-                loss.backward()
+
+                train_losses = self._loss(theta_batch, x_batch, num_atoms)
+                train_loss = torch.mean(train_losses)
+                train_log_probs_sum -= train_losses.sum().item()
+
+                train_loss.backward()
                 if clip_max_norm is not None:
                     clip_grad_norm_(
                         self._neural_net.parameters(),
@@ -224,20 +225,25 @@ class RatioEstimator(NeuralInference, ABC):
 
             self.epoch += 1
 
+            train_log_prob_average = train_log_probs_sum / (
+                len(train_loader) * train_loader.batch_size  # type: ignore
+            )
+            self._summary["train_log_probs"].append(train_log_prob_average)
+
             # Calculate validation performance.
             self._neural_net.eval()
-            loss_sum = 0
+            val_log_prob_sum = 0
             with torch.no_grad():
                 for batch in val_loader:
                     theta_batch, x_batch = (
                         batch[0].to(self._device),
                         batch[1].to(self._device),
                     )
-                    loss = self._loss(theta_batch, x_batch, num_atoms)
-                    loss_sum -= loss.sum().item()
+                    val_losses = self._loss(theta_batch, x_batch, num_atoms)
+                    val_log_prob_sum -= val_losses.sum().item()
                 # Take mean over all validation samples.
-                self._val_log_prob = loss_sum / (
-                    len(val_loader) * val_loader.batch_size
+                self._val_log_prob = val_log_prob_sum / (
+                    len(val_loader) * val_loader.batch_size  # type: ignore
                 )
                 # Log validation log prob for every epoch.
                 self._summary["validation_log_probs"].append(self._val_log_prob)
@@ -262,88 +268,12 @@ class RatioEstimator(NeuralInference, ABC):
         if show_train_summary:
             print(self._describe_round(self._round, self._summary))
 
+        # Avoid keeping the gradients in the resulting network, which can
+        # cause memory leakage when benchmarking.
+        self._neural_net.zero_grad(set_to_none=True)
+
         return deepcopy(self._neural_net)
 
-    def build_posterior(
-        self,
-        density_estimator: Optional[TorchModule] = None,
-        sample_with: str = "mcmc",
-        mcmc_method: str = "slice_np",
-        mcmc_parameters: Optional[Dict[str, Any]] = None,
-        rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
-    ) -> RatioBasedPosterior:
-        r"""
-        Build posterior from the neural density estimator.
-
-        SNRE trains a neural network to approximate likelihood ratios, which in turn
-        can be used obtain an unnormalized posterior
-        $p(\theta|x) \propto p(x|\theta) \cdot p(\theta)$. The posterior returned here
-        wraps the trained network such that one can directly evaluate the unnormalized
-        posterior log-probability $p(\theta|x) \propto p(x|\theta) \cdot p(\theta)$ and
-        draw samples from the posterior with MCMC. Note that, in the case of
-        single-round SNRE_A / AALR, it is possible to evaluate the log-probability of
-        the **normalized** posterior, but sampling still requires MCMC.
-
-        Args:
-            density_estimator: The density estimator that the posterior is based on.
-                If `None`, use the latest neural density estimator that was trained.
-            sample_with: Method to use for sampling from the posterior. Must be one of
-                [`mcmc` | `rejection`].
-            mcmc_method: Method used for MCMC sampling, one of `slice_np`, `slice`,
-                `hmc`, `nuts`. Currently defaults to `slice_np` for a custom numpy
-                implementation of slice sampling; select `hmc`, `nuts` or `slice` for
-                Pyro-based sampling.
-            mcmc_parameters: Dictionary overriding the default parameters for MCMC.
-                The following parameters are supported: `thin` to set the thinning
-                factor for the chain, `warmup_steps` to set the initial number of
-                samples to discard, `num_chains` for the number of chains,
-                `init_strategy` for the initialisation strategy for chains; `prior` will
-                draw init locations from prior, whereas `sir` will use
-                Sequential-Importance-Resampling using `init_strategy_num_candidates`
-                to find init locations.
-            rejection_sampling_parameters: Dictionary overriding the default parameters
-                for rejection sampling. The following parameters are supported:
-                `proposal` as the proposal distribtution (default is the prior).
-                `max_sampling_batch_size` as the batchsize of samples being drawn from
-                the proposal at every iteration. `num_samples_to_find_max` as the
-                number of samples that are used to find the maximum of the
-                `potential_fn / proposal` ratio. `num_iter_to_find_max` as the number
-                of gradient ascent iterations to find the maximum of that ratio. `m` as
-                multiplier to that ratio.
-
-        Returns:
-            Posterior $p(\theta|x)$  with `.sample()` and `.log_prob()` methods
-            (the returned log-probability is unnormalized).
-        """
-
-        if density_estimator is None:
-            density_estimator = self._neural_net
-            # If internal net is used device is defined.
-            device = self._device
-        else:
-            # Otherwise, infer it from the device of the net parameters.
-            device = next(density_estimator.parameters()).device.type
-
-        self._posterior = RatioBasedPosterior(
-            method_family=self.__class__.__name__.lower(),
-            neural_net=density_estimator,
-            prior=self._prior,
-            x_shape=self._x_shape,
-            sample_with=sample_with,
-            mcmc_method=mcmc_method,
-            mcmc_parameters=mcmc_parameters,
-            rejection_sampling_parameters=rejection_sampling_parameters,
-            device=device,
-        )
-
-        self._posterior._num_trained_rounds = self._round + 1
-
-        # Store models at end of each round.
-        self._model_bank.append(deepcopy(self._posterior))
-        self._model_bank[-1].net.eval()
-
-        return deepcopy(self._posterior)
-
     def _classifier_logits(self, theta: Tensor, x: Tensor, num_atoms: int) -> Tensor:
         """Return logits obtained through classifier forward pass.
 
@@ -368,3 +298,105 @@ class RatioEstimator(NeuralInference, ABC):
     @abstractmethod
     def _loss(self, theta: Tensor, x: Tensor, num_atoms: int) -> Tensor:
         raise NotImplementedError
+
+    def build_posterior(
+        self,
+        density_estimator: Optional[nn.Module] = None,
+        prior: Optional[Distribution] = None,
+        sample_with: str = "mcmc",
+        mcmc_method: str = "slice_np",
+        vi_method: str = "rKL",
+        mcmc_parameters: Dict[str, Any] = {},
+        vi_parameters: Dict[str, Any] = {},
+        rejection_sampling_parameters: Dict[str, Any] = {},
+    ) -> Union[MCMCPosterior, RejectionPosterior, VIPosterior]:
+        r"""Build posterior from the neural density estimator.
+
+        SNRE trains a neural network to approximate likelihood ratios. The
+        posterior wraps the trained network such that one can directly evaluate the
+        unnormalized posterior log probability $p(\theta|x) \propto p(x|\theta) \cdot
+        p(\theta)$ and draw samples from the posterior with MCMC or rejection sampling.
+        Note that, in the case of single-round SNRE_A / AALR, it is possible to
+        evaluate the log-probability of the **normalized** posterior, but sampling
+        still requires MCMC (or rejection sampling).
+
+        Args:
+            density_estimator: The density estimator that the posterior is based on.
+                If `None`, use the latest neural density estimator that was trained.
+            prior: Prior distribution.
+            sample_with: Method to use for sampling from the posterior. Must be one of
+                [`mcmc` | `rejection` | `vi`].
+            mcmc_method: Method used for MCMC sampling, one of `slice_np`, `slice`,
+                `hmc`, `nuts`. Currently defaults to `slice_np` for a custom numpy
+                implementation of slice sampling; select `hmc`, `nuts` or `slice` for
+                Pyro-based sampling.
+            vi_method: Method used for VI, one of [`rKL`, `fKL`, `IW`, `alpha`]. Note
+                that some of the methods admit a `mode seeking` property (e.g. rKL)
+                whereas some admit a `mass covering` one (e.g fKL).
+            mcmc_parameters: Additional kwargs passed to `MCMCPosterior`.
+            vi_parameters: Additional kwargs passed to `VIPosterior`.
+            rejection_sampling_parameters: Additional kwargs passed to
+                `RejectionPosterior`.
+
+        Returns:
+            Posterior $p(\theta|x)$  with `.sample()` and `.log_prob()` methods
+            (the returned log-probability is unnormalized).
+        """
+        if prior is None:
+            assert (
+                self._prior is not None
+            ), """You did not pass a prior. You have to pass the prior either at
+                initialization `inference = SNRE(prior)` or to `.build_posterior
+                (prior=prior)`."""
+            prior = self._prior
+        else:
+            check_prior(prior)
+
+        if density_estimator is None:
+            ratio_estimator = self._neural_net
+            # If internal net is used device is defined.
+            device = self._device
+        else:
+            ratio_estimator = density_estimator
+            # Otherwise, infer it from the device of the net parameters.
+            device = next(density_estimator.parameters()).device.type
+
+        potential_fn, theta_transform = ratio_estimator_based_potential(
+            ratio_estimator=ratio_estimator, prior=prior, x_o=None
+        )
+
+        if sample_with == "mcmc":
+            self._posterior = MCMCPosterior(
+                potential_fn=potential_fn,
+                theta_transform=theta_transform,
+                proposal=prior,
+                method=mcmc_method,
+                device=device,
+                x_shape=self._x_shape,
+                **mcmc_parameters,
+            )
+        elif sample_with == "rejection":
+            self._posterior = RejectionPosterior(
+                potential_fn=potential_fn,
+                proposal=prior,
+                device=device,
+                x_shape=self._x_shape,
+                **rejection_sampling_parameters,
+            )
+        elif sample_with == "vi":
+            self._posterior = VIPosterior(
+                potential_fn=potential_fn,
+                theta_transform=theta_transform,
+                prior=prior,  # type: ignore
+                vi_method=vi_method,
+                device=device,
+                x_shape=self._x_shape,
+                **vi_parameters,
+            )
+        else:
+            raise NotImplementedError
+
+        # Store models at end of each round.
+        self._model_bank.append(deepcopy(self._posterior))
+
+        return deepcopy(self._posterior)
