@@ -1,14 +1,9 @@
 # flake8: noqa
-from sbi.utils.conditional_density import (
-    conditional_corrcoeff,
-    eval_conditional_density,
-    extract_and_transform_mog,
-    condition_mog,
-)
+from sbi.utils.conditional_density_utils import extract_and_transform_mog
 from sbi.utils.get_nn_models import classifier_nn, likelihood_nn, posterior_nn
 from sbi.utils.io import get_data_root, get_log_root, get_project_root
 from sbi.utils.kde import KDEWrapper, get_kde
-from sbi.utils.plot import conditional_pairplot, pairplot
+from sbi.utils.potentialutils import pyro_potential_wrapper, transformed_potential
 from sbi.utils.restriction_estimator import RestrictedPrior, RestrictionEstimator
 from sbi.utils.sbiutils import (
     batched_mixture_mv,
@@ -19,14 +14,13 @@ from sbi.utils.sbiutils import (
     del_entries,
     expit,
     get_simulations_since_round,
+    gradient_ascent,
     handle_invalid_x,
     logit,
     mask_sims_from_prior,
+    match_theta_and_x_batch_shapes,
     mcmc_transform,
     mog_log_prob,
-    optimize_potential_fn,
-    rejection_sample,
-    rejection_sample_posterior_within_prior,
     standardizing_net,
     standardizing_transform,
     warn_if_zscoring_changes_data,
@@ -34,6 +28,7 @@ from sbi.utils.sbiutils import (
     warn_on_invalid_x_for_snpec_leakage,
     within_support,
     x_shape_from_simulation,
+    z_score_parser,
 )
 from sbi.utils.torchutils import (
     BoxUniform,
@@ -64,6 +59,8 @@ from sbi.utils.typechecks import (
 )
 from sbi.utils.user_input_checks import (
     check_estimator_arg,
+    check_prior,
+    process_prior,
     process_x,
     test_posterior_net_for_multi_d_x,
     validate_theta_and_x,
