@@ -2,32 +2,53 @@
 # under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.
 
 
-import logging
 import warnings
-from typing import Any, Callable, Optional, Sequence, Tuple, Union, cast
+from typing import Any, Callable, Dict, Optional, Sequence, Tuple, Union, cast
 
 import torch
 from numpy import ndarray
+from pyknos.nflows import flows
 from scipy.stats._distn_infrastructure import rv_frozen
 from scipy.stats._multivariate import multi_rv_frozen
 from torch import Tensor, float32, nn
 from torch.distributions import Distribution, Uniform
 
+from sbi.types import Array
 from sbi.utils.sbiutils import warn_on_iid_x, within_support
 from sbi.utils.torchutils import BoxUniform, atleast_2d
 from sbi.utils.user_input_checks_utils import (
-    CustomPytorchWrapper,
+    CustomPriorWrapper,
     MultipleIndependent,
     PytorchReturnTypeWrapper,
     ScipyPytorchWrapper,
 )
 
 
-def process_prior(prior) -> Tuple[Distribution, int, bool]:
+def check_prior(prior: Any) -> None:
+    """Assert that prior is a PyTorch distribution (or pass if None)."""
+
+    if prior is None:
+        pass
+    else:
+        assert isinstance(
+            prior, Distribution
+        ), """Prior must be a PyTorch Distribution. See FAQ 7 for more details or use
+        `sbi.utils.user_input_checks.process_prior` for wrapping scipy and lists of
+        independent priors."""
+
+
+def process_prior(
+    prior, custom_prior_wrapper_kwargs: Dict = {}
+) -> Tuple[Distribution, int, bool]:
     """Return PyTorch distribution-like prior from user-provided prior.
 
     Args:
         prior: Prior object with `.sample()` and `.log_prob()` as provided by the user.
+        custom_prior_wrapper_kwargs: kwargs to be passed to the class that wraps a
+            custom prior into a pytorch Distribution, e.g., for passing bounds for a
+            prior with bounded support (lower_bound, upper_bound), or argument
+            constraints.
+            (arg_constraints), see pytorch.distributions.Distribution for more info.
 
     Raises:
         AttributeError: If prior objects lacks `.sample()` or `.log_prob()`.
@@ -61,14 +82,21 @@ def process_prior(prior) -> Tuple[Distribution, int, bool]:
 
     # Otherwise it is a custom prior - check for `.sample()` and `.log_prob()`.
     else:
-        return process_custom_prior(prior)
+        return process_custom_prior(prior, custom_prior_wrapper_kwargs)
 
 
-def process_custom_prior(prior) -> Tuple[Distribution, int, bool]:
+def process_custom_prior(
+    prior, custom_prior_wrapper_kwargs: Dict = {}
+) -> Tuple[Distribution, int, bool]:
     """Check and return corrected prior object defined by the user.
 
     Args:
         prior: Prior object with `.sample()` and `.log_prob()` as provided by the user.
+        custom_prior_wrapper_kwargs: kwargs to be passed to the class that wraps a
+            custom prior into a pytorch Distribution, e.g., for passing bounds for a
+            prior with bounded support (lower_bound, upper_bound), or argument
+            constraints.
+            (arg_constraints), see pytorch.distributions.Distribution for more info.
 
     Returns:
         prior: sbi-compatible prior.
@@ -78,18 +106,27 @@ def process_custom_prior(prior) -> Tuple[Distribution, int, bool]:
 
     check_prior_attributes(prior)
     check_prior_batch_behavior(prior)
-    prior, is_prior_numpy = maybe_wrap_prior_as_pytorch(prior)
+    prior, is_prior_numpy = maybe_wrap_prior_as_pytorch(
+        prior, custom_prior_wrapper_kwargs
+    )
     check_prior_return_type(prior)
     theta_numel = prior.sample().numel()
 
     return prior, theta_numel, is_prior_numpy
 
 
-def maybe_wrap_prior_as_pytorch(prior) -> Tuple[Distribution, bool]:
+def maybe_wrap_prior_as_pytorch(
+    prior, custom_prior_wrapper_kwargs: Dict[str, Any] = {}
+) -> Tuple[Distribution, bool]:
     """Check prior return type and maybe wrap as PyTorch.
 
     Args:
         prior: Prior object with `.sample()` and `.log_prob()` as provided by the user.
+        custom_prior_wrapper_kwargs: kwargs to be passed to the class that wraps a
+            custom prior into a pytorch Distribution, e.g., for passing bounds for a
+            prior with bounded support (lower_bound, upper_bound), or argument
+            constraints.
+            (arg_constraints), see pytorch.distributions.Distribution for more info.
 
     Raises:
         TypeError: If prior return type is PyTorch or Numpy.
@@ -106,20 +143,23 @@ def maybe_wrap_prior_as_pytorch(prior) -> Tuple[Distribution, bool]:
     if isinstance(theta, Tensor) and isinstance(log_probs, Tensor):
         # XXX: We wrap to get a Distribution. But this might interfere with the fact
         # that the custom prior can be a probabilistic program.
-        prior = CustomPytorchWrapper(
-            custom_prior=prior, event_shape=torch.Size([theta.numel()])
+        prior = CustomPriorWrapper(
+            custom_prior=prior,
+            event_shape=torch.Size([theta.numel()]),
+            **custom_prior_wrapper_kwargs,
         )
         is_prior_numpy = False
     elif isinstance(theta, ndarray) and isinstance(log_probs, ndarray):
         # infer event shape from single numpy sample.
         event_shape = torch.Size([theta.size])
-        prior = CustomPytorchWrapper(custom_prior=prior, event_shape=event_shape)
+        prior = CustomPriorWrapper(
+            custom_prior=prior, event_shape=event_shape, **custom_prior_wrapper_kwargs
+        )
         is_prior_numpy = True
     else:
         raise TypeError(
             f"Prior must return torch.Tensor or ndarray, but returns {type(theta)}"
         )
-
     return cast(Distribution, prior), is_prior_numpy
 
 
@@ -190,7 +230,7 @@ def check_prior_batch_dims(prior) -> None:
             using `torch.distributions.Independent` to reinterpret batch dimensions as
             event dimensions, or use the `MultipleIndependent` distribution we provide.
 
-            To use `sbi.utils.MultipleIndependent`, just pass a list of priors, e.g. to 
+            To use `sbi.utils.MultipleIndependent`, just pass a list of priors, e.g. to
             specify a uniform prior over two parameters, pass as prior:
                 prior = [
                             Uniform(torch.zeros(1), torch.ones(1)),
@@ -219,8 +259,6 @@ def check_for_possibly_batched_x_shape(x_shape):
 
     x_ndim = len(x_shape)
     inferred_batch_shape, *inferred_data_shape = x_shape
-    # Interpret first dimension as batch dimension.
-    inferred_data_ndim = len(inferred_data_shape)
 
     # Reject multidimensional data with batch_shape > 1.
     if x_ndim > 1 and inferred_batch_shape > 1:
@@ -246,7 +284,7 @@ def check_for_possibly_batched_x_shape(x_shape):
                     of sbi might not provide stable support for this and result in
                     shape mismatches.
 
-            NOTE: below we use list notation to reduce clutter, but `x` should be of 
+            NOTE: below we use list notation to reduce clutter, but `x` should be of
             type torch.Tensor or ndarray.
 
             For example:
@@ -304,7 +342,7 @@ def check_prior_attributes(prior) -> None:
     except:  # Catch any other error.
         raise ValueError(
             f"""Something went wrong when sampling a batch of parameters
-            from the prior as `prior.sample(({num_samples}, ))`. Consider using a 
+            from the prior as `prior.sample(({num_samples}, ))`. Consider using a
             PyTorch distribution."""
         )
     try:
@@ -364,7 +402,7 @@ def check_prior_support(prior):
     """
 
     try:
-        within_support(prior, prior.sample())
+        within_support(prior, prior.sample((1,)))
     except NotImplementedError:
         raise NotImplementedError(
             """The prior must implement the support property or allow to call
@@ -372,6 +410,51 @@ def check_prior_support(prior):
         )
 
 
+def check_embedding_net_device(embedding_net: nn.Module, datum: torch.Tensor) -> None:
+    """Checks if the device for the `embedding_net`'s weights is the same as the device
+    for the fed `datum`. In case of discrepancy, warn the user and move the
+    embedding_net` to  the `datum`'s device.
+
+    Args:
+        embedding_net: torch `Module` embedding data
+        datum torch `Tensor` from the training device
+    """
+    datum_device = datum.device
+    embedding_net_devices = [p.device for p in embedding_net.parameters()]
+    if len(embedding_net_devices) > 0:
+        embedding_net_device = embedding_net_devices[0]
+        if embedding_net_device != datum_device:
+            warnings.warn(
+                "Mismatch between the device of the data fed "
+                "to the embedding_net and the device of the "
+                "embedding_net's weights. Fed data has device "
+                f"'{datum_device}' vs embedding_net weights have "
+                f"device '{embedding_net_device}'. "
+                "Automatically switching the embedding_net's device to "
+                f"'{datum_device}', which could otherwise be done manually "
+                f"""using the line `embedding_net.to('{datum_device}')`."""
+            )
+            embedding_net.to(datum_device)
+    else:
+        pass
+
+
+def check_data_device(datum_1: torch.Tensor, datum_2: torch.Tensor) -> None:
+    """Checks if two tensors have the seme device. Fails if there is a device
+    discrepancy
+
+    Args:
+        datum_1: torch `Tensor`
+        datum_2: torch `Tensor`
+    """
+    assert datum_1.device == datum_2.device, (
+        "Mismatch in fed data's device: "
+        f"datum_1 has device '{datum_1.device}' whereas "
+        f"datum_2 has device '{datum_2.device}'. Please "
+        "use data from a common device."
+    )
+
+
 def process_simulator(
     user_simulator: Callable,
     prior,
@@ -454,9 +537,13 @@ def get_batch_loop_simulator(simulator: Callable) -> Callable:
     return batch_loop_simulator
 
 
-def process_x(x: Tensor, x_shape: torch.Size, allow_iid_x: bool = False) -> Tensor:
+def process_x(
+    x: Array, x_shape: Optional[torch.Size] = None, allow_iid_x: bool = False
+) -> Tensor:
     """Return observed data adapted to match sbi's shape and type requirements.
 
+    If `x_shape` is `None`, the shape is not checked.
+
     Args:
         x: Observed data as provided by the user.
         x_shape: Prescribed shape - either directly provided by the user at init or
@@ -467,7 +554,7 @@ def process_x(x: Tensor, x_shape: torch.Size, allow_iid_x: bool = False) -> Tens
         x: Observed data with shape ready for usage in sbi.
     """
 
-    x = torch.as_tensor(atleast_2d(x), dtype=float32)
+    x = atleast_2d(torch.as_tensor(x, dtype=float32))
 
     input_x_shape = x.shape
     if not allow_iid_x:
@@ -477,12 +564,13 @@ def process_x(x: Tensor, x_shape: torch.Size, allow_iid_x: bool = False) -> Tens
         warn_on_iid_x(num_trials=input_x_shape[0])
         start_idx = 1
 
-    # Number of trials can change for every new x, but single trial x shape must match.
-    assert input_x_shape[start_idx:] == x_shape[start_idx:], (
-        f"Observed data shape ({input_x_shape[start_idx:]}) must match "
-        f"the shape of simulated data x ({x_shape[start_idx:]})."
-    )
-
+    if x_shape is not None:
+        # Number of trials can change for every new x, but single trial x shape must
+        # match.
+        assert input_x_shape[start_idx:] == x_shape[start_idx:], (
+            f"Observed data shape ({input_x_shape[start_idx:]}) must match "
+            f"the shape of simulated data x ({x_shape[start_idx:]})."
+        )
     return x
 
 
@@ -530,7 +618,7 @@ def check_sbi_inputs(simulator: Callable, prior: Distribution) -> None:
     """
     check_prior_support(prior)
     num_prior_samples = 1
-    theta = prior.sample((num_prior_samples,))
+    theta = prior.sample(torch.Size((num_prior_samples,)))
     theta_batch_shape, *_ = theta.shape
     simulation = simulator(theta)
     sim_batch_shape, *sim_event_shape = simulation.shape
@@ -550,16 +638,6 @@ def check_sbi_inputs(simulator: Callable, prior: Distribution) -> None:
 
 def check_estimator_arg(estimator: Union[str, Callable]) -> None:
     """Check (density or ratio) estimator argument passed by the user."""
-    if isinstance(estimator, BoxUniform) or isinstance(estimator, Distribution):
-        raise ValueError(
-            "You passed a distribution as density_estimator. Probably, "
-            "your code is deprecated since sbi v0.14.0, which changed the "
-            "API. Please consult release notes to see how you can update your code: "
-            "https://github.com/mackelab/sbi/releases/tag/v0.14.0 "
-            "More information can be found under the corresponding pull request on "
-            "github: https://github.com/mackelab/sbi/pull/378 and tutorials: "
-            "https://www.mackelab.org/sbi/tutorial/02_flexible_interface/"
-        )
     assert isinstance(estimator, str) or (
         isinstance(estimator, Callable) and not isinstance(estimator, nn.Module)
     ), (
@@ -601,19 +679,26 @@ def validate_theta_and_x(
     assert theta.dtype == float32, "Type of parameters must be float32."
     assert x.dtype == float32, "Type of simulator outputs must be float32."
 
-    simulations_device = f"{x.device.type}:{x.device.index}"
-    if "cpu" not in simulations_device and "cpu" in training_device:
-        logging.warning(
-            f"""Simulations are on {simulations_device} but training device is
-            set to {training_device}, moving data to device to {training_device}."""
+    if str(x.device) != training_device:
+        warnings.warn(
+            f"Data x has device '{x.device}' "
+            f"different from the training_device '{training_device}', "
+            f"moving x to the training_device '{training_device}'."
         )
         x = x.to(training_device)
+
+    if str(theta.device) != training_device:
+        warnings.warn(
+            f"Parameters theta has device '{theta.device}' "
+            f"different from the training_device '{training_device}', "
+            f"moving theta to the training_device '{training_device}'."
+        )
         theta = theta.to(training_device)
 
     return theta, x
 
 
-def test_posterior_net_for_multi_d_x(net: nn.Module, theta: Tensor, x: Tensor) -> None:
+def test_posterior_net_for_multi_d_x(net: flows.Flow, theta: Tensor, x: Tensor) -> None:
     """Test log prob method of the net.
 
     This is done to make sure the net can handle multidimensional inputs via an
