@@ -12,15 +12,25 @@ from torch.distributions import MultivariateNormal
 
 from sbi import analysis as analysis
 from sbi import utils as utils
-
-from sbi.inference import SNPE_A, SNPE_B, SNPE_C, prepare_for_sbi, simulate_for_sbi
-
+from sbi.analysis import ConditionedMDN, conditonal_potential
+from sbi.inference import (
+    SNPE_A,
+    SNPE_B,
+    SNPE_C,
+    DirectPosterior,
+    MCMCPosterior,
+    RejectionPosterior,
+    posterior_estimator_based_potential,
+    prepare_for_sbi,
+    simulate_for_sbi,
+)
 from sbi.simulators.linear_gaussian import (
     linear_gaussian,
     samples_true_posterior_linear_gaussian_mvn_prior_different_dims,
     samples_true_posterior_linear_gaussian_uniform_prior,
     true_posterior_linear_gaussian_mvn_prior,
 )
+from tests.sbiutils_test import conditional_of_mvn
 from tests.test_utils import (
     check_c2st,
     get_dkl_gaussian_prior,
@@ -28,8 +38,6 @@ from tests.test_utils import (
     get_prob_outside_uniform_prior,
 )
 
-from tests.sbiutils_test import conditional_of_mvn
-
 
 @pytest.mark.parametrize("snpe_method", [SNPE_A, SNPE_C])
 @pytest.mark.parametrize(
@@ -44,17 +52,13 @@ from tests.sbiutils_test import conditional_of_mvn
     ),
 )
 def test_c2st_snpe_on_linearGaussian(
-    snpe_method, num_dim: int, prior_str: str, num_trials: int, set_seed
+    snpe_method, num_dim: int, prior_str: str, num_trials: int
 ):
-    """Test whether SNPE infers well a simple example with available ground truth.
-
-    Args:
-        set_seed: fixture for manual seeding
-    """
+    """Test whether SNPE infers well a simple example with available ground truth."""
 
     x_o = zeros(num_trials, num_dim)
     num_samples = 1000
-    num_simulations = 2500
+    num_simulations = 2600
 
     # likelihood_mean will be likelihood_shift+theta
     likelihood_shift = -1.0 * ones(num_dim)
@@ -83,14 +87,18 @@ def test_c2st_snpe_on_linearGaussian(
     theta, x = simulate_for_sbi(
         simulator, prior, num_simulations, simulation_batch_size=1000
     )
-    _ = inference.append_simulations(theta, x).train(training_batch_size=100)
-    posterior = inference.build_posterior().set_default_x(x_o)
+    posterior_estimator = inference.append_simulations(theta, x).train(
+        training_batch_size=100
+    )
+    posterior = DirectPosterior(
+        prior=prior, posterior_estimator=posterior_estimator
+    ).set_default_x(x_o)
     samples = posterior.sample((num_samples,))
 
     # Compute the c2st and assert it is near chance level of 0.5.
     check_c2st(samples, target_samples, alg="snpe_c")
 
-    map_ = posterior.map(num_init_samples=1_000)
+    map_ = posterior.map(num_init_samples=1_000, show_progress_bars=False)
 
     # Checks for log_prob()
     if prior_str == "gaussian":
@@ -121,7 +129,7 @@ def test_c2st_snpe_on_linearGaussian(
             posterior_likelihood_unnorm,
             posterior_likelihood_norm,
             acceptance_prob,
-        ) = get_normalization_uniform_prior(posterior, prior, x_o)
+        ) = get_normalization_uniform_prior(posterior, prior, x=x_o)
         # The acceptance probability should be *exactly* the ratio of the unnormalized
         # and the normalized likelihood. However, we allow for an error margin of 1%,
         # since the estimation of the acceptance probability is random (based on
@@ -135,15 +143,13 @@ def test_c2st_snpe_on_linearGaussian(
         assert ((map_ - ones(num_dim)) ** 2).sum() < 0.5
 
 
-def test_c2st_snpe_on_linearGaussian_different_dims(set_seed):
+def test_c2st_snpe_on_linearGaussian_different_dims():
     """Test whether SNPE B/C infer well a simple example with available ground truth.
 
     This example has different number of parameters theta than number of x. Also
     this implicitly tests simulation_batch_size=1. It also impleictly tests whether the
     prior can be `None` and whether we can stop and resume training.
 
-    Args:
-        set_seed: fixture for manual seeding
     """
 
     theta_dim = 3
@@ -183,10 +189,14 @@ def test_c2st_snpe_on_linearGaussian_different_dims(set_seed):
     theta, x = simulate_for_sbi(simulator, prior, 2000, simulation_batch_size=1)
 
     inference = inference.append_simulations(theta, x)
-    _ = inference.train(max_num_epochs=10)  # Test whether we can stop and resume.
-    _ = inference.train(resume_training=True)
-    posterior = inference.build_posterior()
-    samples = posterior.sample((num_samples,), x=x_o)
+    posterior_estimator = inference.train(
+        max_num_epochs=10
+    )  # Test whether we can stop and resume.
+    posterior_estimator = inference.train(resume_training=True)
+    posterior = DirectPosterior(
+        prior=prior, posterior_estimator=posterior_estimator
+    ).set_default_x(x_o)
+    samples = posterior.sample((num_samples,))
 
     # Compute the c2st and assert it is near chance level of 0.5.
     check_c2st(samples, target_samples, alg="snpe_c")
@@ -197,6 +207,7 @@ def test_c2st_snpe_on_linearGaussian_different_dims(set_seed):
 @pytest.mark.parametrize(
     "method_str",
     (
+        "snpe_a",
         pytest.param(
             "snpe_b",
             marks=pytest.mark.xfail(
@@ -205,14 +216,11 @@ def test_c2st_snpe_on_linearGaussian_different_dims(set_seed):
         ),
         "snpe_c",
         "snpe_c_non_atomic",
-        "snpe_a",
     ),
 )
-def test_c2st_multi_round_snpe_on_linearGaussian(method_str: str, set_seed):
+def test_c2st_multi_round_snpe_on_linearGaussian(method_str: str):
     """Test whether SNPE B/C infer well a simple example with available ground truth.
-
-    Args:
-        set_seed: fixture for manual seeding.
+    .
     """
 
     num_dim = 2
@@ -233,7 +241,10 @@ def test_c2st_multi_round_snpe_on_linearGaussian(method_str: str, set_seed):
     target_samples = gt_posterior.sample((num_samples,))
 
     if method_str == "snpe_c_non_atomic":
-        density_estimator = utils.posterior_nn("mdn", num_components=5)
+        # Test whether SNPE works properly with structured z-scoring.
+        density_estimator = utils.posterior_nn(
+            "mdn", z_score_x="structured", num_components=5
+        )
         method_str = "snpe_c"
     elif method_str == "snpe_a":
         density_estimator = "mdn_snpe_a"
@@ -244,7 +255,6 @@ def test_c2st_multi_round_snpe_on_linearGaussian(method_str: str, set_seed):
         lambda theta: linear_gaussian(theta, likelihood_shift, likelihood_cov), prior
     )
     creation_args = dict(
-        simulator=simulator,
         prior=prior,
         density_estimator=density_estimator,
         show_progress_bars=False,
@@ -253,21 +263,28 @@ def test_c2st_multi_round_snpe_on_linearGaussian(method_str: str, set_seed):
     if method_str == "snpe_b":
         inference = SNPE_B(**creation_args)
         theta, x = simulate_for_sbi(simulator, prior, 500, simulation_batch_size=10)
-        _ = inference.append_simulations(theta, x).train()
-        posterior1 = inference.build_posterior().set_default_x(x_o)
+        posterior_estimator = inference.append_simulations(theta, x).train()
+        posterior1 = DirectPosterior(
+            prior=prior, posterior_estimator=posterior_estimator
+        ).set_default_x(x_o)
         theta, x = simulate_for_sbi(
             simulator, posterior1, 1000, simulation_batch_size=10
         )
-        _ = inference.append_simulations(theta, x, proposal=posterior1).train()
-        posterior = inference.build_posterior().set_default_x(x_o)
+        posterior_estimator = inference.append_simulations(
+            theta, x, proposal=posterior1
+        ).train()
+        posterior = DirectPosterior(
+            prior=prior, posterior_estimator=posterior_estimator
+        ).set_default_x(x_o)
     elif method_str == "snpe_c":
         inference = SNPE_C(**creation_args)
-        theta, x = simulate_for_sbi(simulator, prior, 500, simulation_batch_size=50)
-        _ = inference.append_simulations(theta, x).train()
-        posterior1 = inference.build_posterior().set_default_x(x_o)
-        theta, x = simulate_for_sbi(
-            simulator, posterior1, 1000, simulation_batch_size=50
-        )
+        theta, x = simulate_for_sbi(simulator, prior, 900, simulation_batch_size=50)
+        posterior_estimator = inference.append_simulations(theta, x).train()
+        posterior1 = DirectPosterior(
+            prior=prior, posterior_estimator=posterior_estimator
+        ).set_default_x(x_o)
+        theta = posterior1.sample((1000,))
+        x = simulator(theta)
         _ = inference.append_simulations(theta, x, proposal=posterior1).train()
         posterior = inference.build_posterior().set_default_x(x_o)
     elif method_str == "snpe_a":
@@ -304,12 +321,10 @@ def test_c2st_multi_round_snpe_on_linearGaussian(method_str: str, set_seed):
         ("rejection", "rejection", "uniform"),
     ),
 )
-def test_api_snpe_c_posterior_correction(sample_with, mcmc_method, prior_str, set_seed):
+def test_api_snpe_c_posterior_correction(sample_with, mcmc_method, prior_str):
     """Test that leakage correction applied to sampling works, with both MCMC and
     rejection.
 
-    Args:
-        set_seed: fixture for manual seeding
     """
 
     num_dim = 2
@@ -329,29 +344,36 @@ def test_api_snpe_c_posterior_correction(sample_with, mcmc_method, prior_str, se
     simulator, prior = prepare_for_sbi(
         lambda theta: linear_gaussian(theta, likelihood_shift, likelihood_cov), prior
     )
-    inference = SNPE_C(
-        prior,
-        simulation_batch_size=50,
-        sample_with=sample_with,
-        mcmc_method=mcmc_method,
-        show_progress_bars=False,
-    )
+    inference = SNPE_C(prior, show_progress_bars=False)
 
     theta, x = simulate_for_sbi(simulator, prior, 1000)
-    _ = inference.append_simulations(theta, x).train(max_num_epochs=5)
-
-    posterior = inference.build_posterior()
-    posterior = posterior.set_sample_with(sample_with).set_mcmc_method(mcmc_method)
+    posterior_estimator = inference.append_simulations(theta, x).train()
+    potential_fn, theta_transform = posterior_estimator_based_potential(
+        posterior_estimator, prior, x_o
+    )
+    if sample_with == "mcmc":
+        posterior = MCMCPosterior(
+            potential_fn=potential_fn,
+            theta_transform=theta_transform,
+            proposal=prior,
+            method=mcmc_method,
+        )
+    elif sample_with == "rejection":
+        posterior = RejectionPosterior(
+            potential_fn=potential_fn,
+            proposal=prior,
+            theta_transform=theta_transform,
+        )
 
     # Posterior should be corrected for leakage even if num_rounds just 1.
-    samples = posterior.sample((10,), x=x_o)
+    samples = posterior.sample((10,))
 
     # Evaluate the samples to check correction factor.
-    posterior.log_prob(samples, x=x_o)
+    _ = posterior.log_prob(samples)
 
 
 @pytest.mark.slow
-def test_sample_conditional(set_seed):
+def test_sample_conditional():
     """
     Test whether sampling from the conditional gives the same results as evaluating.
 
@@ -379,7 +401,8 @@ def test_sample_conditional(set_seed):
         else:
             return linear_gaussian(theta, -likelihood_shift, likelihood_cov)
 
-    net = utils.posterior_nn("maf", hidden_features=20)
+    # Test whether SNPE works properly with structured z-scoring.
+    net = utils.posterior_nn("maf", z_score_x="structured", hidden_features=20)
 
     simulator, prior = prepare_for_sbi(simulator, prior)
 
@@ -387,20 +410,38 @@ def test_sample_conditional(set_seed):
 
     # We need a pretty big dataset to properly model the bimodality.
     theta, x = simulate_for_sbi(simulator, prior, 10000)
-    _ = inference.append_simulations(theta, x).train(max_num_epochs=50)
+    posterior_estimator = inference.append_simulations(theta, x).train(
+        max_num_epochs=60
+    )
 
-    posterior = inference.build_posterior().set_default_x(x_o)
+    posterior = DirectPosterior(
+        prior=prior, posterior_estimator=posterior_estimator
+    ).set_default_x(x_o)
     samples = posterior.sample((50,))
 
     # Evaluate the conditional density be drawing samples and smoothing with a Gaussian
     # kde.
-    cond_samples = posterior.sample_conditional(
-        (500,), condition=samples[0], dims_to_sample=[dim_to_sample_1, dim_to_sample_2]
+    potential_fn, theta_transform = posterior_estimator_based_potential(
+        posterior_estimator, prior=prior, x_o=x_o
     )
+    (conditioned_potential_fn, restricted_tf, restricted_prior,) = conditonal_potential(
+        potential_fn=potential_fn,
+        theta_transform=theta_transform,
+        prior=prior,
+        condition=samples[0],
+        dims_to_sample=[dim_to_sample_1, dim_to_sample_2],
+    )
+    mcmc_posterior = MCMCPosterior(
+        potential_fn=conditioned_potential_fn,
+        theta_transform=restricted_tf,
+        proposal=restricted_prior,
+    )
+    cond_samples = mcmc_posterior.sample((500,))
+
     _ = analysis.pairplot(
         cond_samples,
         limits=[[-2, 2], [-2, 2], [-2, 2]],
-        fig_size=(2, 2),
+        figsize=(2, 2),
         diag="kde",
         upper="kde",
     )
@@ -432,10 +473,9 @@ def test_sample_conditional(set_seed):
     error = np.abs(sample_kde_grid - eval_grid.numpy())
 
     max_err = np.max(error)
-    assert max_err < 0.0026
+    assert max_err < 0.0027
 
 
-@pytest.mark.slow
 def test_mdn_conditional_density(num_dim: int = 3, cond_dim: int = 1):
     """Test whether the conditional density infered from MDN parameters of a
     `DirectPosterior` matches analytical results for MVN. This uses a n-D joint and
@@ -454,7 +494,7 @@ def test_mdn_conditional_density(num_dim: int = 3, cond_dim: int = 1):
 
     x_o = zeros(1, num_dim)
     num_samples = 1000
-    num_simulations = 2500
+    num_simulations = 2700
     condition = 0.1 * ones(1, num_dim)
 
     dims = list(range(num_dim))
@@ -486,17 +526,18 @@ def test_mdn_conditional_density(num_dim: int = 3, cond_dim: int = 1):
         return linear_gaussian(theta, likelihood_shift, likelihood_cov)
 
     simulator, prior = prepare_for_sbi(simulator, prior)
-    inference = SNPE_C(prior, show_progress_bars=False, density_estimator="mdn")
+    inference = SNPE_C(density_estimator="mdn", show_progress_bars=False)
 
     theta, x = simulate_for_sbi(
         simulator, prior, num_simulations, simulation_batch_size=1000
     )
-    _ = inference.append_simulations(theta, x).train(training_batch_size=100)
-    posterior = inference.build_posterior().set_default_x(x_o)
-
-    conditional_samples_sbi = posterior.sample_conditional(
-        (num_samples,), condition, dims2sample, x_o
+    posterior_mdn = inference.append_simulations(theta, x).train(
+        training_batch_size=100
     )
+    conditioned_mdn = ConditionedMDN(
+        posterior_mdn, x_o, condition=condition, dims_to_sample=[0]
+    )
+    conditional_samples_sbi = conditioned_mdn.sample((num_samples,))
     check_c2st(
         conditional_samples_sbi,
         conditional_samples_gt,
@@ -526,12 +567,15 @@ def test_example_posterior(snpe_method: type):
     simulator, prior = prepare_for_sbi(
         lambda theta: linear_gaussian(theta, likelihood_shift, likelihood_cov), prior
     )
-    inference = snpe_method(prior, show_progress_bars=False, **extra_kwargs)
+    inference = snpe_method(prior, show_progress_bars=False)
 
     theta, x = simulate_for_sbi(
         simulator, prior, 1000, simulation_batch_size=10, num_workers=6
     )
-    _ = inference.append_simulations(theta, x).train()
-
-    posterior = inference.build_posterior().set_default_x(x_o)
+    posterior_estimator = inference.append_simulations(theta, x).train(**extra_kwargs)
+    if snpe_method == SNPE_A:
+        posterior_estimator = inference.correct_for_proposal()
+    posterior = DirectPosterior(
+        prior=prior, posterior_estimator=posterior_estimator
+    ).set_default_x(x_o)
     assert posterior is not None
